@@ -7,7 +7,7 @@
  *
  * Description: Newt UI for cdebconf
  *
- * $Id: newt.c 57544 2009-02-13 16:44:52Z cjwatson $
+ * $Id: newt.c 59368 2009-07-16 15:18:17Z cjwatson $
  *
  * cdebconf is (c) 2000-2001 Randolph Chung and others under the following
  * license.
@@ -117,8 +117,16 @@ struct newtColors newtAltColorPalette = {
 
 typedef int (newt_handler)(struct frontend *obj, struct question *q);
 
+struct newt_help_callback_data {
+    struct frontend *obj;
+    const char *tag;
+};
+
 static void newt_progress_stop(struct frontend *obj);
 
+/* Result must be freed by the caller */
+static char *get_full_description(struct frontend *obj, struct question *q);
+
 #include "cdebconf_newt.h"
 
 /*  Padding of title width, allows for leading "[!!] " before title
@@ -126,6 +134,14 @@ static void newt_progress_stop(struct frontend *obj);
  */
 #define TITLE_PADDING 9
 
+/*  At least one blue line on top
+ *  One line for the title / top frame
+ *  One line for the bottom frame
+ *  One line for shadow on the bottom
+ *  One line with a menu
+ */
+#define MIN_DECORATION_HEIGHT 5
+
 /* gettext would be much nicer :-( */
 static const char *
 continue_text(struct frontend *obj)
@@ -160,7 +176,13 @@ cancel_text(struct frontend *obj)
 static const char *
 help_text(struct frontend *obj)
 {
-    return question_get_text(obj, "debconf/help-line", "<Tab> moves between items; <Space> selects; <Enter> activates buttons");
+    return question_get_text(obj, "debconf/help-line", "<Tab> moves; <Space> selects; <Enter> activates buttons");
+}
+
+static const char *
+help_text_f1(struct frontend *obj)
+{
+    return question_get_text(obj, "debconf/help-line-f1", "<F1> for help; <Tab> moves; <Space> selects; <Enter> activates buttons");
 }
 
 void
@@ -185,6 +207,17 @@ cdebconf_newt_get_progress_info(struct frontend *obj)
         return NULL;
 }
 
+/* This is global due to cdebconf_newt_create_form's interface. It might be
+ * worth fixing this at the next convenient API break.
+ */
+static struct newt_help_callback_data *help_cb_data;
+
+newtComponent
+cdebconf_newt_create_form(newtComponent scrollbar)
+{
+    return newtForm(scrollbar, help_cb_data, 0);
+}
+
 void
 cdebconf_newt_create_window(const int width, const int height, const char *title, const char *priority)
 {
@@ -275,15 +308,27 @@ cdebconf_newt_get_text_width(const char *text)
 static int
 min_window_height(struct frontend *obj, struct question *q, int win_width)
 {
+    // start with a blank or description (note and error)
+    // End with <Continue>/boolean buttons + blank
     int height = 3;
     char *type = q->template->type;
-    char *q_ext_text;
+    char *q_text;
 
-    q_ext_text = q_get_extended_description(obj, q);
-    if (q_ext_text != NULL)
-        height = cdebconf_newt_get_text_height(q_ext_text, win_width) + 1;
-    if (strcmp(type, "multiselect") == 0 || strcmp(type, "select") == 0)
-        height += 4; // at least three lines for choices + blank line
+    if (strcmp(q->template->type, "note") == 0 || strcmp(q->template->type, "error") == 0)
+        q_text = q_get_extended_description(obj, q);
+    else
+        q_text = get_full_description(obj, q);
+    if (q_text != NULL) {
+        height += cdebconf_newt_get_text_height(q_text, win_width) + 1;
+        free (q_text);
+    }
+    if (strcmp(type, "multiselect") == 0)
+        height += 4; // x lines for choices + blank line
+    else if (strcmp(type, "select") == 0) {
+        height += 2; // as multiselect, but without a button and blank line
+        if (obj->methods.can_go_back(obj, q))
+            height += 2;
+    }
     else if (strcmp(type, "string") == 0 || strcmp(type, "password") == 0)
         height += 2; // input line + blank line
     // the others don't need more space
@@ -298,7 +343,7 @@ need_separate_window(struct frontend *obj, struct question *q)
 
     newtGetScreenSize(&width, &height);
     x = min_window_height(obj, q, width-7);
-    return (x >= height-5);
+    return (x > height-MIN_DECORATION_HEIGHT);
 }
 
 static char *
@@ -323,7 +368,7 @@ get_full_description(struct frontend *obj, struct question *q)
 }
 
 static int
-show_separate_window(struct frontend *obj, struct question *q)
+show_separate_window(struct frontend *obj, struct question *q, int is_help)
 {
     newtComponent form, textbox, bOk, bCancel, cRet;
     int width = 80, height = 24, t_height, t_width, win_width, win_height;
@@ -380,7 +425,7 @@ show_separate_window(struct frontend *obj, struct question *q)
     if (t_width_descr > t_width)
         t_width = t_width_descr;
     t_width_buttons = 2*BUTTON_PADDING + cdebconf_newt_get_text_width(continue_text(obj)) + 2;
-    if (obj->methods.can_go_back(obj, q))
+    if (!is_help && obj->methods.can_go_back(obj, q))
         //  Add an interspace
         t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
@@ -398,7 +443,7 @@ show_separate_window(struct frontend *obj, struct question *q)
     assert(textbox);
     newtTextboxSetText(textbox, full_description);
     free(full_description);
-    if (obj->methods.can_go_back(obj, q)) {
+    if (!is_help && obj->methods.can_go_back(obj, q)) {
         bOk     = newtCompactButton(win_width - TEXT_PADDING - BUTTON_PADDING - strwidth(continue_text(obj)) - 3, win_height-2, continue_text(obj));
         bCancel = newtCompactButton(TEXT_PADDING + BUTTON_PADDING - 1,  win_height-2, goback_text(obj));
         newtFormAddComponents(form, bCancel, textbox, bOk, NULL);
@@ -676,6 +721,8 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
     int listflags = NEWT_FLAG_RETURNEXIT;
     int width = 80, height = 24;
     int win_width, win_height = -1, t_height, t_width, sel_height, sel_width;
+    int select_list_top;
+    int b_height;
     int t_width_title, t_width_buttons;
     char **choices, **choices_trans, *defval;
     int count = 0, i, ret, defchoice = -1;
@@ -755,18 +802,28 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
         t_height = newtTextboxGetNumLines(textbox);
         newtTextboxSetHeight(textbox, t_height);
         newtFormAddComponent(form, textbox);
-    } else
+        select_list_top = 1+t_height+1;
+    } else {
         t_height = 0;
+        select_list_top = 1; // No description. Only insert a blank line.
+    }
+    if (obj->methods.can_go_back(obj, q))
+        b_height = 2;
+    else
+        b_height = 0;
     free(full_description);
-    win_height = t_height + 5 + sel_height;
-    if (win_height > height-5) {
-        win_height = height-5;
-        sel_height = win_height - t_height - 5;
+    win_height  = t_height + sel_height + b_height;
+    //    3 == First blank line + blanks before and after select
+    // or 3 == Blanks before and after select + blank after <Go Back>
+    win_height += 3;
+    if (win_height > height-MIN_DECORATION_HEIGHT) {
+        win_height = height-MIN_DECORATION_HEIGHT;
+        sel_height = win_height - t_height - 3 - b_height;
     }
     if (count > sel_height)
         listflags |= NEWT_FLAG_SCROLL;
     cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
-    listbox = newtListbox((win_width-sel_width-3)/2, 1+t_height+1, sel_height, listflags);
+    listbox = newtListbox((win_width-sel_width-3)/2, select_list_top, sel_height, listflags);
     defval = (char *)question_getvalue(q, "");
     for (i = 0; i < count; i++) {
         newtListboxAppendEntry(listbox, choices_trans[i], choices[tindex[i]]);
@@ -917,7 +974,7 @@ newt_handler_multiselect(struct frontend *obj, struct question *q)
     separate_window = need_separate_window(obj, q);
     while (1) {
         if (separate_window) {
-            ret = show_separate_window(obj, q);
+            ret = show_separate_window(obj, q, 0);
             if (ret != DC_OK)
                 return ret;
             ret = show_multiselect_window(obj, q, 0);
@@ -945,7 +1002,7 @@ newt_handler_select(struct frontend *obj, struct question *q)
     separate_window = need_separate_window(obj, q);
     while (1) {
         if (separate_window) {
-            ret = show_separate_window(obj, q);
+            ret = show_separate_window(obj, q, 0);
             if (ret != DC_OK)
                 return ret;
             ret = show_select_window(obj, q, 0);
@@ -995,7 +1052,7 @@ static int newt_handler_password(struct frontend *obj, struct question *q)
 static int newt_handler_note(struct frontend *obj, struct question *q)
 {
     // Sort of hack-ish, yes, but it works :-)
-    return show_separate_window(obj, q);
+    return show_separate_window(obj, q, 0);
 }
 
 /*
@@ -1039,6 +1096,23 @@ newt_handler_error(struct frontend *obj, struct question *q)
     return ret;
 }
 
+static void
+newt_help_callback(newtComponent co, void *d)
+{
+    struct newt_help_callback_data *data = d;
+    struct frontend *obj = data->obj;
+    struct question *q = obj->qdb->methods.get(obj->qdb, data->tag);
+
+    /* Don't allow recursive calls. */
+    newtSetHelpCallback(NULL);
+
+    show_separate_window(obj, q, 1);
+
+    newtSetHelpCallback(newt_help_callback);
+
+    question_deref(q);
+}
+
 /* ----------------------------------------------------------------------- */
 struct question_handlers {
 	const char *type;
@@ -1123,21 +1197,37 @@ newt_go(struct frontend *obj)
             }
 
             if (plugin || strcmp(q->template->type, question_handlers[i].type) == 0) {
+                char *help = q_get_help(obj, q);
+                struct newt_help_callback_data my_help_cb_data;
                 if (!cleared && !data->scale_form) {
                     cleared = 1;
                     cdebconf_newt_setup();
                 }
+                if (*help) {
+                    my_help_cb_data.obj = obj;
+                    my_help_cb_data.tag = help;
+                    help_cb_data = &my_help_cb_data;
+                    newtSetHelpCallback(newt_help_callback);
+                }
                 if (obj->info != NULL) {
                     char *text = q_get_description(obj, obj->info);
                     if (text)
                         newtDrawRootText(0, 0, text);
                     free(text);
                 }
-                newtPushHelpLine(help_text(obj));
+                if (*help)
+                    newtPushHelpLine(help_text_f1(obj));
+                else
+                    newtPushHelpLine(help_text(obj));
                 ret = handler(obj, q);
                 newtPopHelpLine();
                 if (ret == DC_OK)
                     obj->qdb->methods.set(obj->qdb, q);
+                if (*help) {
+                    newtSetHelpCallback(NULL);
+                    help_cb_data = NULL;
+                }
+                free(help);
                 if (plugin)
                     plugin_delete(plugin);
                 break;
